/***************************************************************************/
/*
 * atp.c --
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * $Id: atp.c,v 1.8 95/06/01 12:16:36 tfisher Exp Locker: tfisher $
 *
 * For more information:
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include "atp.h"
#include "atp_api.h"


/*
 * Exported Routines.
 * ------------------
 */
int atp_open(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	char *peer);		/* hostname of the remote communication peer */
int atp_send(
	int sessID,		/* session ID returned from atp_open */
	char *buff,		/* pointer to the buffer of data to send */
	int size);		/* size of the buffer of data to send */
int atp_receive(
	int sessID,		/* session ID returned from atp_open */
	char *buff,		/* pointer to the buffer to receive data */
	int size);		/* size of the buffer to receive data */
int atp_close(
	int sessID);		/* session ID returned from atp_open */
void atp_print_send_stats(
	int sessID);		/* session ID returned from atp_open */
void atp_print_receive_stats(
	int sessID);		/* session ID returned from atp_open */


/*
 * Local Routines.
 * ---------------
 */
static int atp_open_myrinet(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	char *peer);		/* hostname of the remote communication peer */
static int atp_send_myrinet(
	struct session *sess,	/* pointer to unique session struct */
	char *buff,		/* buffer of data to send */
	int size);		/* size of the buffer of data to send */
static int atp_send_myrinet_ACK(
	struct session *sess,	/* pointer to unique session struct */
	struct packet *pkt,	/* pointer to packet header to process */
	struct packet *ack_pkt); /* pointer to packet header to send */
static int atp_receive_myrinet(
	struct session *sess,	/* pointer to unique session struct */
	char *buff,		/* buffer to receive data */
	int size);		/* size of the buffer to receive data */
static int atp_close_myrinet(
	struct session *sess);	/* pointer to unique session struct */
static int atp_copy_packet(
	struct packet *pkt,	/* pointer to packet header to copy */
	char *data,		/* pointer to packet data to copy */
	struct buffer *buff);	/* buffer to be copied into */
static int atp_send_packet(
	struct session *sess,	/* pointer to unique session struct */
	struct packet *pkt,	/* pointer to packet header to send */
	char *data);		/* pointer to packet data to send */
static struct buffer *atp_get_send_buff(
	struct session *sess);	/* pointer to unique session struct */
static void atp_add_receive_buff(
	struct session *sess,	/* pointer to unique session struct */
	struct buffer *buff);	/* buffer to be added to the LANai */
static struct session *sess_alloc(void);
static struct session *sess_access(
	int sessID);		/* session ID returned from atp_open */
static int sess_free(
	struct session *sess);	/* pointer to unique session struct */


/*
 * DEBUG Routines.
 * ---------------
 */
static inline void dbugSend(
	int size, 
	int window, 
	int retrans, 
	struct buffer *buff, 
	struct packet *pkt);
static inline void dbugSendPing(
	int size,
	struct packet *pkt);
static inline void dbugReceive(
	int size, 
	int window, 
	int retrans, 
	struct buffer *buff, 
	struct packet *pkt);
static inline void dbugReceivePingAck(
	struct packet *ack_pkt);
static inline void dbugReceiveData(
	int npacket, 
	int nbundle, 
	int size, 
	int total, 
	struct packet *pkt);
static inline void dbugReceiveTimeOut(
	int xmitno, 
	int count, 
	int window, 
	struct packet *ack_pkt);
static inline void dbugSendAck(
	struct packet *ack_pkt);
static inline void dbugReceiveAck(
	struct packet *ack_pkt);
static inline void dbugReceiveAckTimeOut();
static inline void dbugReceiveZero(
	int count, 
	int window, 
	struct packet *ack_pkt);
static inline void dbugReceiveStale(
	int xmit_seq_no,
	int nwindow,
	int count,
	struct packet *pkt,
	struct buffer *buff);
static inline void dbugReceiveDuplicate(
	int size, 
	int total, 
	struct packet *pkt);
static inline void dbugReceiveAckDuplicate(
	struct packet *pkt,
	struct packet *ack_pkt);
static inline void dbugReceiveFinUnexpected(
	struct packet *ack_pkt);
static inline void dbugReceiveCopy(
	int npacket, 
	int nbundle, 
	struct buffer *buff, 
	struct packet *pkt);
static inline void dbugReceiveCheck(
	int count, 
	int window, 
	int retrans);
static inline void dbugResizeWindow(
	int count, 
	int window);
static inline void dbugPacketSize(
	int oddsize, 
	int count, 
	int last);
static inline void dbugSendPrepRetrans(
	int window, 
	struct packet *ack_pkt);
static inline void dbugReceivePrepRetrans(
	int window, 
	int new_window, 
	struct packet *ack_pkt);
static inline void dbugUnreachable(
	char *hostname);


/*
 * Global variables.
 * -----------------
 */
struct session *myrinet_sessions[MAX_SESSIONS];


/*
 * atp_open() --
 * Initializes reliable and ordered packet communication.  
 */
int
atp_open(
	int unit,
	int interface,
	char *peer)
{
	return (atp_open_myrinet(unit, interface, peer));
}

/*
 * atp_close() --
 */
int
atp_close(
	int sessID)
{
	int rval;
	struct session *sess;
	char *fn = "atp_close";

	if (sessID > MAX_SESSIONS)
		Error(1, "%s: Bad sessID.", fn);
	if ((sess = sess_access(sessID)) == 0)
		Error(1, "%s: No active session.", fn);

	rval = atp_close_myrinet(sess);
	return (rval);
}

/*
 * atp_send() --
 * Verifies the session is valid.
 * Sends a buffer of unlimited size.
 */
int
atp_send(
	int sessID,
	char *buff,
	int size)
{
	int rval;
	struct session *sess;
	char *fn = "atp_send";

	if (sessID > MAX_SESSIONS)
		Error(1, "%s: Bad sessID.", fn);
	if ((sess = sess_access(sessID)) == 0)
		Error(1, "%s: No active session.", fn);

	rval = atp_send_myrinet(sess, buff, size);
	return (rval);
}

/*
 * atp_receive() --
 * Verifies the session is valid.
 * Receives a buffer of unlimited size.
 */
int
atp_receive(
	int sessID,
	char *buff,
	int size)
{
	int rval;
	struct session *sess;
	char *fn = "atp_receive";

	if (sessID > MAX_SESSIONS)
		Error(1, "%s: Bad sessID.", fn);
	if ((sess = sess_access(sessID)) == 0)
		Error(1, "%s: No active session.", fn);

	rval = atp_receive_myrinet(sess, buff, size);
	return (rval);
}

static int
atp_open_myrinet(
	int unit,
	int interface,
	char *peer)
{
	struct session *sess;
	char *fn = "atp_open_myrinet";

	/* Allocate a session. */
	sess = sess_alloc();
	sess->unit = unit;
	sess->interface = interface;
	sess->proto = ATP_MYRINET;
	sess->mtu = MAX_MTU;
	sess->rtt = (struct timeval) {0, 0};

	/* Initialize libAtpApi and the session. */
	AtomicInit(sess->unit, sess->interface, MYRINET_CP, sess->mtu);
	sess->nbuffs = AtomicOpen(sess->unit, sess->interface, 
	    &sess->ID, MAX_BUFFS, sess->buff);

	/* Get the destination host ID */
	if ((sess->addr_peer = AtomicGetIdFromName(peer)) == 0)
		Error(1, "[%d] %s: Couldn't get peer %s hostID.", 
		    getpid(), fn, peer);
	return (sess->ID);
}

/*
 * atp_close_myrinet() --
 */
static int
atp_close_myrinet(
	struct session *sess)
{
	AtomicClose(sess->unit, sess->interface, sess->ID);
	sess_free(sess);
	return (0);
}

/*
 * atp_send_myrinet() --
 * 	This routine first sends a message to the specific destination host 
 * thereby requesting an immediate response so as to determine its operational 
 * status and its approximate distance away so that an appropriate timeout 
 * value can be used for determining when to retransmit unacknowledged packets. 
 * 	We wait "sess->rtt" (the estimated round trip time) amount of time 
 * for an ACK of each "window"-sized number of packets sent.  If we receive no 
 * ACK then we incease the time to wait by twice and resend all the packets 
 * of the previous window.  If we still get no response then we ping the dest-
 * ination host to a new "sess->rtt" and try to send again.  Otherwise if 
 * we get no response from the ping, we give up.
 */
static int
atp_send_myrinet(
	struct session *sess,
	char *data,
	int size)
{
	int i, cc;
	int tmp=0;
	int pos=0;
	int finished=0;		/* indicates completion of transmission */
	int count=0;		/* packet count sent within window */
	int window=0;		/* dynamically changing window size */
	int retries=0;
	int nnacks=0;
	int copied;
	int missing;
	int last=0;
	int offset=0;
	int retrans=0;
	int total=0;		/* total amount of data sent */
	int oddsize;
	int nwindow=0;
	int npacket=0;
	struct packet pkt;
	struct packet ack_pkt;
	struct buffer *buff;
	struct stats *send_stats;
	struct timeval timeout;
	static int xmit_seq_no=0;
	char *fn = "atp_send_myrinet";

	xmit_seq_no++;
	timeout.tv_sec = 0;
	timeout.tv_usec = 500000;
	send_stats = &sess->send_stats;
	send_stats->n_sessions++;

	/*
	 * Ping the receiver.  
	 * First, initialize the packet header and then adjust the data 
	 * size to be a multiple of 4 if it isn't already.  If the size of 
	 * the data is then still less than the MTU (we are only sending 
	 * one packet), then send it with the ping and immediately listen 
	 * for the ACK which should also be a FIN.  In this case we bypass 
	 * the larger send loop (where fragmentation takes place) further 
	 * below and instead loop until the "ping" has been ACK'ed and exit 
	 * after receiving the ACK.
	 */
	pkt.pkt_size = 0;
	pkt.pkt_seq = 0;
	pkt.pkt_bun = 0;
	pkt.pkt_xmit = xmit_seq_no;
	pkt.pkt_flags = PKT_SYN;
	/*XXX while (size % 4)
		size++;*/
	if (size <= MAX_MSG) {
		pkt.pkt_size = size;
		pkt.pkt_flags |= PKT_DAT | PKT_FIN;
		send_stats->n_segments++;
		finished++;
	}
handshake:
	for (;;) {
		/*
		 * Send the packet.
		 */
		if (atp_send_packet(sess, &pkt, data) <= 0)
			continue;
		dbugSendPing(size, &pkt);
		send_stats->n_ping_sends++;

		/*
		 * Receive the ACK.
		 * If time-out here then we "continue" back up to the
		 * top of the loop and retransmit the original packet.
		 * Or, return zero if we have exercised all of our retries.
		 */
		for (;;) {
			cc = AtomicReceive(sess->unit, 
			    sess->interface, &buff, &timeout);
			if (cc == 0) {
				send_stats->n_ping_timeouts++;
				dbugReceiveAckTimeOut();
				if (retries++ == MAX_RECV_ACK_RETRY) {
					dbugUnreachable(sess->host_peer);
					return (0);
				} else
				    goto handshake;
			} else if (cc < 0) {
				Error(0, "%s: No ACK.", fn);
				continue;
			} else if (cc < PACKET_SIZE) {
				send_stats->n_corrupted_acks++;
				Error(0, "%s: Corrupted ACK, size=%d.", fn, cc);
				continue;
			}
			bcopy(buff->u, (char *)&ack_pkt, PACKET_SIZE);
			atp_add_receive_buff(sess, buff);

			/* Check for duplicate ACKs. */
			if (ack_pkt.pkt_xmit != pkt.pkt_xmit) {
				dbugReceiveAckDuplicate(&pkt, &ack_pkt);
				send_stats->n_duplicate_acks++;
				continue;
			}
			break;
		}

		/*
		 * Process the ACK.
		 * Return here if ACK has a FIN.
		 */
		if ((ack_pkt.pkt_flags & PKT_SYN) &&
		    (ack_pkt.pkt_flags & PKT_ACK)) {
			dbugReceivePingAck(&ack_pkt);
			send_stats->n_ping_acks++;
			if (ack_pkt.pkt_flags & PKT_FIN) {
				debug("DONE.\n");
				return (pkt.pkt_size);
			}
			break;
		} else {
			send_stats->n_corrupted_acks++;
			Error(0, ">>> Unexpected ACK: flags=0x%x.\n", 
			    ack_pkt.pkt_flags);
		}
	}

	/*
	 * Initializations prior to large packet transfers (below).
	 */
	oddsize = 0;
	ack_pkt.pkt_nacks = -1;
	window = (size / MAX_MSG) + 1;
	if (window > MAX_WINDOW)
		window = MAX_WINDOW;

sendloop:
	/*
	 * Send loop.
	 * (XXX Need to check the return value from AtomicSend()).
	 * The interleaving of data copies with DMA operations occurs within
	 * this loop.
	 */
	copied = 0;
	missing = 0;
	pkt.pkt_flags = PKT_DAT;
	for (count=0, finished=0, pos=1; 
	     count < MAX_WINDOW && !finished; pos *= 2, count++) {
		if ((ack_pkt.pkt_nacks & pos) == 0)
			continue;

		/*
		 * Figure out the packet's size, and decrement bytes 
		 * left.  If retransmitting the last packet of the window 
		 * (count+1 == MAX_WINDOW) and "oddsize" is set then we are
		 * retransmitting an oddsized packet which also requires
		 * FIN flag set.
		 */
		if (retrans) {
			if (++missing == window) {
				if (oddsize && count == last) {
					pkt.pkt_size = oddsize;
					pkt.pkt_flags |= PKT_FIN;
				} else {
					pkt.pkt_size = MAX_MSG;
					dbugPacketSize(oddsize, count, last);
				}
			} else
			    pkt.pkt_size = MAX_MSG;
			send_stats->n_retransmits++;
		} else if (size > MAX_MSG) {
			pkt.pkt_size = MAX_MSG;
			size -= pkt.pkt_size;
			send_stats->n_segments++;
		} else {
			oddsize = size;
			last = count;
			pkt.pkt_size = size;
			size -= pkt.pkt_size;
			pkt.pkt_flags |= PKT_FIN;
			send_stats->n_segments++;
		}

		/*
		 * Copy the data to the DMA buffer.
		 */
		pkt.pkt_seq = count;
		pkt.pkt_bun = nwindow;
		buff = atp_get_send_buff(sess);
		atp_copy_packet(&pkt, data, buff);
		send_stats->n_data_copies++;

		/* 
		 * Send the packet asynchronously.
		 */
		cc = AtomicSend(sess->unit, sess->interface, 
		    sess->addr_peer, buff, pkt.pkt_size+PACKET_SIZE);
		dbugSend(size, window, retrans, buff, &pkt);
		send_stats->n_sends++;
		if (!retrans)
			total += pkt.pkt_size;
		if (pkt.pkt_flags & PKT_FIN)
			finished++;
		if (retrans && missing == window)
			break;
	}

	/*
	 * Receive the ACK.
	 */
	for (;;) {
		cc = AtomicReceive(sess->unit, 
		    sess->interface, &buff, (struct timeval *)-1);
		if (cc <= 0) {
			debug(">>> %s: Resending All.\n", fn);
			send_stats->n_resends++;
			ack_pkt.pkt_nacks = -1;
			retrans = 1;
			goto sendloop;
		}
		if (cc > 0) {
			if (cc != PACKET_SIZE) {
				send_stats->n_corrupted_acks++;
				Error(0, "%s: Corrupted ACK.", fn);
				continue;
			}
			bcopy(buff->u, (char *)&ack_pkt, PACKET_SIZE);
			atp_add_receive_buff(sess, buff);

			/*
			 * Process the ACK.
			 */
			if (ack_pkt.pkt_flags & PKT_SYN) {
				Error(0, "%s: Received a SYN.\n", fn); 
				continue;
			}
			if ((ack_pkt.pkt_flags & PKT_ACK) == 0) {
				send_stats->n_corrupted_acks++;
				Error(1, "%s: Received packet not an ACK.", fn);
				continue;
			}
			dbugReceiveAck(&ack_pkt);
			send_stats->n_window_acks++;
			if (ack_pkt.pkt_flags & PKT_FIN)
				return (total);
			if (ack_pkt.pkt_flags & PKT_NACK) {
				dbugSendPrepRetrans(window, &ack_pkt);
				send_stats->n_negative_acks++;
				window = ack_pkt.pkt_win;
				retrans = 1;
				nnacks++;
				goto sendloop;
			}
			nwindow++;
			retrans = 0;
			if (finished) {
				send_stats->n_missing_fins++;
				Error(0, "%s: No expected FIN.\n", fn); 
			}
			ack_pkt.pkt_nacks = -1;
			goto sendloop;
		}
	}
}

/*
 * atp_receive_myrinet() --
 * Verifies that all received packets are from the expected sender, if such
 * packets are from some other source then they are dropped thus enforcing
 * protection against malicious senders.  This protection is further enhanced
 * by the fact that the sending LANai verifies internally the source address 
 * field of each link level packet header.
 */
int
atp_receive_myrinet(
	struct session *sess,
	char *data,
	int size)
{
	int i;
	int cc;			/* character count received */
	int started=0;		/* indicates transmission setup complete */
	int finished=0;		/* indicates completion of transmission */
	int count=0;		/* packet count sent within window */
	int position=0;		/* bit position within NACK mask of ACK */
	int window;		/* dynamic packet receive window size */
	int orig_window;	/* static packet receive window size */
	int nnacks=0;		/* count of retransmission requests */
	int total=0;		/* total amount unduplicated data received */
	int retries;
	int tmp=0;
	int offset=0;
	int retrans=0;
	int missing;
	int nwindow=0;
	int nbundle=0;
	int npacket=0;
	int proto_time_xxx;
	struct packet pkt;
	struct packet ack_pkt;
	struct buffer *buff;
	struct stats *recv_stats;
	struct timeval timeout;
	struct timeval start;
	struct timeval finish;
	struct timeval elapsed;
	static int xmit_seq_no=0;
	char *fn = "atp_receive_myrinet";

	/*
	 * Initializations before the receive loop.
	 */
	xmit_seq_no++;
	ack_pkt.pkt_nacks = -1;
	recv_stats = &sess->recv_stats;
	recv_stats->n_sessions++;
	timeout.tv_sec = 0;
	timeout.tv_usec = 50000;

recvloop:
	/*
	 * Receive loop.
	 * (XXX Need to handle the case where buff->size != pkt->pkt_size).
	 * Here at the top we first figure out the window size for the
	 * expected amount of data to be received.  We do this before the
	 * loop always during regular communication -- not during retrans-
	 * missions since we already know how many retransmissions we are
	 * expecting.
	 */
	count = 0;
	position = 1;
	missing = 0;
	retries = 0;
	if (!retrans) {
		dbugResizeWindow(count, window);
		window = (size / MAX_MSG) + 1;
		if (window > MAX_WINDOW)
			window = MAX_WINDOW;
	}
	while (count < MAX_WINDOW) {
		if (retrans && 
		    (ack_pkt.pkt_nacks & position) == 0) {
			count++;
			position *= 2;
			continue;
		}

		/*
		 * Receive data or just listen.
		 */
		cc = AtomicReceive(sess->unit, 
		    sess->interface, &buff, &timeout);

		if (cc == 0) {
			/*atp_print(".");*/
			dbugReceiveTimeOut(xmit_seq_no, 
			    nwindow, count, &ack_pkt);
			recv_stats->n_window_timeouts++;
			if (++retries >= MAX_RECV_DAT_RETRY) {
#ifdef XXX
				ack_pkt.pkt_seq = 0;
				ack_pkt.pkt_bun = nbundle;
				ack_pkt.pkt_win = window;
				ack_pkt.pkt_xmit = xmit_seq_no;
				if (pkt.pkt_flags & PKT_FIN)
					ack_pkt.pkt_flags |= PKT_FIN;
				atp_send_myrinet_ACK(sess, &pkt, &ack_pkt);
				recv_stats->n_window_acks++;
#endif
				recv_stats->n_retries++;
				retries = 0;
				if (npacket)
				    break;
			}
			position *= 2;
			continue;
		} else if (cc < 0) {
			dbugReceiveZero(count, window, &ack_pkt);
			recv_stats->n_zerolen++;
			retries++;
			continue;
		} else if (buff->queue != RECEIVE_QUEUE) {
			recv_stats->n_unexpected_recvq++;
			Error(1, "%s: Unexpected receive queue %d.", 
			    fn, buff->queue);
		}

#ifdef PROTO_TIME
gettimeofday(&start, 0);
proto_time_xxx = 0;
#endif
		/* 
		 * A packet has been received.
		 */
		bcopy(buff->u, (char *)&pkt, PACKET_SIZE);
		recv_stats->n_receives++;

		/*
		 * Verify that the packet is valid.
		 */
		if (pkt.pkt_xmit != xmit_seq_no ||
		    pkt.pkt_bun != nwindow) {
			dbugReceiveStale(xmit_seq_no, 
			    nwindow, count, &pkt, buff);
			recv_stats->n_stales++;
			atp_add_receive_buff(sess, buff);
			proto_time_xxx++;
			continue;
		}

		/*
		 * Packet is valid.
		 * Process the packet header.
		 */
		if (!retrans)
		    npacket++;
		else
		    recv_stats->n_retransmits++;
		dbugReceive(size,
		    window, retrans, buff, &pkt);
		if (pkt.pkt_flags & PKT_SYN) {
			if (started) {
				proto_time_xxx++;
				continue;
			}
			started++;
			ack_pkt.pkt_flags = PKT_SYN;
		}
		if (pkt.pkt_flags & PKT_DAT) {
			if ((cc -= PACKET_SIZE) != pkt.pkt_size) {
				recv_stats->n_unexpected_sizes++;
				Error(0, "%s: Expected size %d, got %d.",
				    fn, cc, pkt.pkt_size);
				atp_add_receive_buff(sess, buff);
				proto_time_xxx++;
				continue;
			}
			if ((ack_pkt.pkt_nacks & 
			    (u_long)exp2((double)pkt.pkt_seq)) == 0) {
				dbugReceiveDuplicate(size, total, &pkt);
				recv_stats->n_duplicates++;
				proto_time_xxx++;
				continue;
			}
			ack_pkt.pkt_nacks &= 
			    ~(u_long)exp2((double)pkt.pkt_seq);
			recv_stats->n_segments++;
			total += pkt.pkt_size;
			size -= pkt.pkt_size;
			position *= 2;
			count++;
		}
		if (pkt.pkt_flags & PKT_FIN) {
			ack_pkt.pkt_flags |= PKT_FIN;
			finished++;
		}

		/*
		 * Send the Ping ACK.
		 * (XXX Need a send loop here with MAX_RETRYS).
		 */
		if (pkt.pkt_flags & PKT_SYN) {
			ack_pkt.pkt_seq = 0;
			ack_pkt.pkt_bun = 0;
			ack_pkt.pkt_xmit = xmit_seq_no;
			if (pkt.pkt_flags & PKT_FIN) {
				ack_pkt.pkt_flags |= PKT_FIN;
				if (size != 0)
				    dbugReceiveFinUnexpected(&ack_pkt);
			}
			atp_send_myrinet_ACK(sess, &pkt, &ack_pkt);
			recv_stats->n_ping_acks++;
			proto_time_xxx++;
		}

		/*
		 * Copy the data out of the DMA buffer. We do the copy of 
		 * the received data during the above DMA of the ACK. There 
		 * is an interesting calculation of the offset here. There 
		 * really should be one variable here being incremented 
		 * instead of having this expensive calculation. Next, we 
		 * mark the buff as ready to use and give it back to the LCP.
		 */
		offset = (pkt.pkt_seq * MAX_MSG) +
		    (MAX_WINDOW * MAX_MSG * nbundle);
		dbugReceiveCopy(npacket, nbundle, buff, &pkt);
		bcopy(buff->u + PACKET_SIZE, (char *)data + offset,
		    pkt.pkt_size);
		atp_add_receive_buff(sess, buff);
		recv_stats->n_data_copies++;

		/* 
		 * Break out of this current receive loop if/when a packet 
		 * with the FIN flag set arrives.  Also, check if/when our
		 * size variable (which gives the expected number of bytes
		 * still yet to be received) reaches zero (or less) that 
		 * the expected FIN flag is also set.
		 */
		if (size == 0 && !finished)
			Print("%s: Unexpected size reduction.\n", fn);
		if (retrans && ++missing == window)
			break;
		if (pkt.pkt_flags & PKT_FIN) {
			if (pkt.pkt_flags & PKT_SYN)
				return (total);
			break;
		}
		proto_time_xxx++;
	}

	/*
	 * Check if we need to send a NACK and thus attempt to receive 
	 * any dropped packets.  If so we adjust the window size to match 
	 * exactly the number of packets that need retransmitting. And 
	 * we send this "window update" along with the NACK.
	 */
	if (!retrans)
		orig_window = window;
	dbugReceiveCheck(count, window, retrans);
	if (ack_pkt.pkt_nacks & ((u_long)exp2((double)orig_window)-1)) {
		int new_window = 0;
		ack_pkt.pkt_size = 0;
		ack_pkt.pkt_flags |= PKT_NACK;
		for (i=0, position=1; i < orig_window; position *= 2, i++)
			if (ack_pkt.pkt_nacks & position)
				new_window++;
		dbugReceivePrepRetrans(
		    orig_window, new_window, &ack_pkt);
		recv_stats->n_negative_acks++;
		window = new_window;
		retrans = 1;
		nnacks++;
	} else if (finished)
		ack_pkt.pkt_flags |= PKT_FIN;

	/*
	 * Send the ACK.
	 * (XXX Need a send loop here with MAX_RETRYS).
	 * Copy the received data during DMA of the ACK.
	 */
	if ((count == window) || 
	    (ack_pkt.pkt_flags & PKT_FIN) ||
	    (ack_pkt.pkt_flags & PKT_NACK) || retrans) {
		ack_pkt.pkt_seq = 0;
		ack_pkt.pkt_bun = nbundle;
		ack_pkt.pkt_win = window;
		ack_pkt.pkt_xmit = xmit_seq_no;
		if (pkt.pkt_flags & PKT_FIN) {
			ack_pkt.pkt_flags |= PKT_FIN;
			if (size != 0)
			    dbugReceiveFinUnexpected(&ack_pkt);
		}
#ifdef PROTO_TIME
gettimeofday(&finish, 0);
#endif
		atp_send_myrinet_ACK(sess, &pkt, &ack_pkt);
		recv_stats->n_window_acks++;

#ifdef PROTO_TIME
if (proto_time_xxx == 0) {
    timersub(&finish, &start, &elapsed);
    Print(">>> %03d useconds.\n", elapsed.tv_usec);
}
#endif
	}

	/*
	 * If expecting retransmissions then go back to the start.
	 */
	if (ack_pkt.pkt_flags & PKT_NACK) {
		ack_pkt.pkt_flags = 0;
		goto recvloop;
	}

	/*
	 * Continue looping (receiving) if not finished.
	 * Resize the window based on the amount of data left to be received.
	 */
	nwindow++;
	retrans = 0;
	if (!finished) {
		ack_pkt.pkt_nacks = -1;
		nbundle++;
		goto recvloop;
	}

	/* 
	 * All data received, we're done.
	 */ 
	debug("DONE.\n");
	return (total);
}

static void
timersub(
	struct timeval *t0,
	struct timeval *t1,
	struct timeval *out)
{
	timerclear(out);
	out->tv_sec = t0->tv_sec - t1->tv_sec;
	out->tv_usec = t0->tv_usec - t1->tv_usec;
	if (t0->tv_usec < t1->tv_usec) {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
}

static int
atp_copy_packet(
	struct packet *pkt,
	char *data,
	struct buffer *buff)
{
	bcopy((char *)pkt, buff->u, PACKET_SIZE);
	if (pkt->pkt_size && data)
	    bcopy((char *)data, buff->u+PACKET_SIZE, pkt->pkt_size);
	return (0);
}

static int
atp_send_packet(
	struct session *sess,
	struct packet *pkt,
	char *data)
{
	int cc;
	int tmp=0;
	int queue;
	struct buffer *buff;
	char *fn = "atp_send_packet";

	/* 
	 * Small packet flow.
	 * Send it via the "immediate" functionality.
	 * If this packet is an ACK then we must use the ACK queue.
	 */
	if (pkt->pkt_size == 0 && pkt->pkt_flags & PKT_ACK) {
		queue = ((pkt->pkt_flags & PKT_ACK) ? 
		    ACK_QUEUE : RECEIVE_QUEUE);
		cc = AtomicSendImmediate(sess->unit, sess->interface, 
		    sess->addr_peer, queue, (char *)pkt, PACKET_SIZE);
		return (cc);
	}

	/* 
	 * Normal packet flow.
	 * First get the next avalaible DMA send buffer.  
	 * Then copy the header and data into it and send it via the 
	 * normal Atomic send functionality.
	 */
	buff = atp_get_send_buff(sess);
	atp_copy_packet(pkt, data, buff);
	cc = AtomicSend(sess->unit, sess->interface, sess->addr_peer, 
	    buff, pkt->pkt_size+PACKET_SIZE);
	return (cc);
}


/*
 * atp_send_myrinet_ACK() --
 * Send an acknowledgment to the sender reflecting our receive status (that
 * is which packets we have successfully received).  If sending an ACK here 
 * if/when nothing has yet been received then the pkt->pkt_seq will still be 
 * uninitialized.
 */
static int
atp_send_myrinet_ACK(
	struct session *sess,
	struct packet *pkt,
	struct packet *ack_pkt)
{
	int cc;

	ack_pkt->pkt_size = 0;
	ack_pkt->pkt_flags |= PKT_ACK;
	if (ack_pkt->pkt_flags & PKT_NACK)
		ack_pkt->pkt_flags &= ~PKT_FIN;	
	dbugSendAck(ack_pkt);
	cc = AtomicSendImmediate(sess->unit, sess->interface,
	    sess->addr_peer, ACK_QUEUE, (char *)ack_pkt, PACKET_SIZE);
	if (ack_pkt->pkt_flags & PKT_NACK)
	    ack_pkt->pkt_flags = PKT_NACK;
	else
	    ack_pkt->pkt_flags = 0;
	return (cc);
}

/* 
 * atp_get_send_buff() --
 * Get the next available DMA send buffer.  Loop through the current set 
 * of DMA send buffers for this session and then return a pointer to 
 * the next one that is marked as owned by "us" (as opposed to being
 * owned by the interface board).  The chosen buffer is also assumed to 
 * be the most recently used one.
 */
static struct buffer *
atp_get_send_buff(
	struct session *sess)
{
	struct buffer *buff;
	static int index=0;

	for (;;) {
		index++;
		index %= sess->nbuffs;
		buff = sess->buff[index];
		if (buff->owned)
			return (buff);
		/* Clear unfinished ACKs. */
		while (AtomicGetSendAck(sess->unit, sess->interface, 0, -1))
			;
	}
}

/*
 * atp_add_receive_buff() --
 * Mark the buff as ready to use and give it back to the LANai control 
 * program.
 */
static void
atp_add_receive_buff(
	struct session *sess,
	struct buffer *buff)
{
	int rval;
	char *fn = "atp_add_receive_buff";

	if ((rval = AtomicAddReceiveBuff(
	    sess->unit, sess->interface, 
	    buff->index, (u_char *)buff->d, 
	    (u_short)buff->size, buff->queue)) == 0) {
		debug("[%d] %s: Can't add buff-%d to queue-%d.",
		    getpid(), fn, buff->index, buff->queue);
	} else if (rval == -1) {
		debug("[%d] %s: LANai reset.\n", 
		    getpid(), fn);
	} else {
		buff->owned = 0;
		/*debug("[%d] %s: LANai added receive buff-%d to queue-%d.\n",
		    getpid(), fn, buff->index, buff->queue);*/
	}
}

/*
 * atp_print_send_stats() --
 */
void
atp_print_send_stats(
	int sessID)		/* session ID returned from atp_open */
{
	struct session *sess;
	char *fn = "atp_print_send_stats";

	if (sessID > MAX_SESSIONS)
		Error(1, "%s: Bad sessID.", fn);
	if ((sess = sess_access(sessID)) == 0)
		Error(1, "%s: No active session.", fn);

#ifdef XXX
	printf("Myrinet SEND stats for %d transactions.\n",
	    sess->send_stats.n_sessions);
	printf("n_segments = %d.\n",
	    sess->send_stats.n_segments);
	printf("n_sends = %d.\n", 
	    sess->send_stats.n_sends);
#endif
	if (sess->send_stats.n_retries)
		printf("n_timeouts = %d.\n", 
		    sess->send_stats.n_retries);
	if (sess->send_stats.n_retransmits)
		printf("n_retransmits = %d.\n", 
		    sess->send_stats.n_retransmits);
	if (sess->send_stats.n_duplicates)
		printf("n_duplicates = %d.\n", 
		    sess->send_stats.n_duplicates);
#ifdef XXX
	printf("n_data_copies = %d.\n", 
	    sess->send_stats.n_data_copies);
	printf("n_resends = %d.\n", 
	    sess->send_stats.n_resends);
	printf("n_ping_timeouts = %d.\n", 
	    sess->send_stats.n_ping_timeouts);
	printf("n_window_timeouts = %d.\n", 
	    sess->send_stats.n_window_timeouts);
	printf("n_zerolen_acks = %d.\n", 
	    sess->send_stats.n_zerolen);
	printf("n_window_acks = %d.\n", 
	    sess->send_stats.n_window_acks);
	printf("n_negative_acks = %d.\n", 
	    sess->send_stats.n_negative_acks);
	printf("n_duplicate_acks = %d.\n", 
	    sess->send_stats.n_duplicate_acks);
	printf("n_corrupted_acks = %d.\n", 
	    sess->send_stats.n_corrupted_acks);
	printf("n_unexpected_recvq = %d.\n", 
	    sess->send_stats.n_unexpected_recvq);
	printf("n_unexpected_sizes = %d.\n", 
	    sess->send_stats.n_unexpected_sizes);
	printf("n_missing_fins = %d.\n", 
	    sess->send_stats.n_missing_fins);
#endif
}

/*
 * atp_print_send_stats() --
 */
void
atp_print_receive_stats(
	int sessID)		/* session ID returned from atp_open */
{
	struct session *sess;
	char *fn = "atp_print_receive_stats";

	if (sessID > MAX_SESSIONS)
		Error(1, "%s: Bad sessID.", fn);
	if ((sess = sess_access(sessID)) == 0)
		Error(1, "%s: No active session.", fn);

#ifdef XXX
	printf("Myrinet RECEIVE stats for %d sessions.\n",
	    sess->recv_stats.n_sessions);
	printf("n_segments = %d.\n",
	    sess->recv_stats.n_segments);
	printf("n_receives = %d.\n", 
	    sess->recv_stats.n_receives);
#endif
	if (sess->recv_stats.n_retries)
		printf("n_timeouts = %d.\n", 
		    sess->recv_stats.n_retries);
	if (sess->recv_stats.n_retransmits)
		printf("n_retransmits = %d.\n", 
		    sess->recv_stats.n_retransmits);
	if (sess->recv_stats.n_duplicates)
		printf("n_duplicates = %d.\n", 
		    sess->recv_stats.n_duplicates);
#ifdef XXX
	printf("n_data_copies = %d.\n", 
	    sess->recv_stats.n_data_copies);
	printf("n_resends = %d.\n", 
	    sess->recv_stats.n_resends);
	printf("n_timeouts = %d.\n", 
	    sess->recv_stats.n_window_timeouts);
	printf("n_retries = %d.\n", 
	    sess->recv_stats.n_retries);
	printf("n_duplicates = %d.\n", 
	    sess->recv_stats.n_duplicates);
	printf("n_stales = %d.\n", 
	    sess->recv_stats.n_stales);
	printf("n_singles = %d.\n", 
	    sess->recv_stats.n_singles);
	printf("n_zerolen = %d.\n", 
	    sess->recv_stats.n_zerolen);
	printf("n_window_acks = %d.\n", 
	    sess->recv_stats.n_window_acks);
	printf("n_negative_acks = %d.\n", 
	    sess->recv_stats.n_negative_acks);
	printf("n_duplicate_acks = %d.\n", 
	    sess->recv_stats.n_duplicate_acks);
	printf("n_corrupted_acks = %d.\n", 
	    sess->recv_stats.n_corrupted_acks);
	printf("n_unexpected_recvq = %d.\n", 
	    sess->recv_stats.n_unexpected_recvq);
	printf("n_unexpected_sizes = %d.\n", 
	    sess->recv_stats.n_unexpected_sizes);
	printf("n_missing_fins = %d.\n", 
	    sess->recv_stats.n_missing_fins);
#endif
}

static struct session *
sess_alloc()
{
	int i;
	struct session *sess;

	if ((sess = (struct session *)
	    malloc((u_int)SESSION_SIZE)) == 0)
		Error(1, "sess_alloc: malloc() failed.");
	bzero((char *)sess, SESSION_SIZE);
	for (i=0; i < MAX_SESSIONS; i++)
		if (myrinet_sessions[i] == 0) {
			myrinet_sessions[i] = sess;
			break;
		}
	if (i == MAX_SESSIONS)
		Error(1, "sess_alloc: No more available.");
	return (sess);
}

static struct session *
sess_access(int sessID)
{
	int i;
	struct session *sess;

	for (i=0; i < MAX_SESSIONS; i++) {
		sess = myrinet_sessions[i];
		if (sess && sess->ID == sessID)
			return (sess);
	}
	return (0);
}

static int
sess_free(struct session *sess)
{
	myrinet_sessions[sess->ID] = 0;
	free((char *)sess);
	return (0);
}


/* 
 * DEBUG Routines from this point on:
 * ----------------------------------
 */

static inline void
dbugSendPing(
	int size,
	struct packet *pkt)
{
#ifdef DEBUG
	Print(">>> Sending Ping=%d:%d: size=%d, cc=%d.\n", pkt->pkt_xmit, pkt->pkt_seq, size, pkt->pkt_size);
#endif
}

static inline void
dbugSend(
	int size, 
	int window, 
	int retrans, 
	struct buffer *buff, 
	struct packet *pkt)
{
#ifdef DEBUG
	/*if (retrans)*/
	Print("Sending %d:%d:%d: size=%d, cc=%d, flags=0x%x, window=%d, buff=%d.\n", pkt->pkt_xmit, pkt->pkt_bun, pkt->pkt_seq, size, pkt->pkt_size, pkt->pkt_flags, window, buff->index);
#endif
}

static inline void
dbugReceivePingAck(
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print(">>> Received Ping ACK=%d: flags=0x%x, nacks=0x%x.\n", ack_pkt->pkt_xmit, ack_pkt->pkt_flags, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceive(
	int size,
	int window,
	int retrans,
	struct buffer *buff, 
	struct packet *pkt)
{
#ifdef DEBUG
	/*if (retrans)*/
		Print("Received=%d:%d:%d, size=%d, cc=%d, flags=0x%x, win=%d, buff=%d\n", pkt->pkt_xmit, pkt->pkt_bun, pkt->pkt_seq, size, pkt->pkt_size, pkt->pkt_flags, window, buff->index);
#endif
}

static inline void
dbugReceiveData(
	int npacket, 
	int nbundle, 
	int size, 
	int total, 
	struct packet *pkt)
{
#ifdef DEBUGXXX
	if (retrans)
		Print("Retrans DAT: bundle=%d:%d, flags=0x%x, size=%d, limit=%d, total=%d\n", nbundle, pkt->pkt_seq, pkt->pkt_flags, pkt->pkt_size, size, total);
	else
		Print("Received DAT=%d: bundle=%d:%d, flags=0x%x, size=%d, limit=%d, total=%d\n", npacket, nbundle, pkt->pkt_seq, pkt->pkt_flags, pkt->pkt_size, size, total);
#endif
}

static inline void
dbugSendAck(
	struct packet *ack_pkt)
{
#ifdef DEBUG
	/*if (ack_pkt->pkt_flags & PKT_NACK)*/
		Print(">>> Sending ACK=%d:%d: flags=0x%x, nacks=0x%x.\n", ack_pkt->pkt_xmit, ack_pkt->pkt_bun, ack_pkt->pkt_flags, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceiveAck(
	struct packet *ack_pkt)
{
#ifdef DEBUG
	/*if (ack_pkt.pkt_flags & PKT_NACK)*/
	Print(">>> Received ACK=%d: flags=0x%x, nacks=0x%x\n", ack_pkt->pkt_xmit, ack_pkt->pkt_flags, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceiveTimeOut(
	int xmitno, 
	int nwindow, 
	int count, 
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("  Timed-out=%d:%d:%d: flags=0x%x, nacks=0x%x.\n", xmitno, nwindow, count, ack_pkt->pkt_flags, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceiveAckTimeOut()
{
#ifdef DEBUG
	Print("ACK-receipt timed-out.\n");
#endif
}

static inline void
dbugReceiveZero(
	int count, 
	int window, 
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("Zero-length packet: count=%d, window=%d, flags=0x%x, nacks=0x%x.\n", count, window, ack_pkt->pkt_flags, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceiveStale(
	int xmit_seq_no,
	int nwindow,
	int count,
	struct packet *pkt,
	struct buffer *buff)
{
#ifdef DEBUG
	Print("Stale receipt=%d:%d:%d, expected=%d:%d:%d, buff=%d.\n", pkt->pkt_xmit, pkt->pkt_bun, pkt->pkt_seq, xmit_seq_no, nwindow, count, buff->index);
#endif
}

static inline void
dbugReceiveDuplicate(
	int size, 
	int total, 
	struct packet *pkt)
{
#ifdef DEBUG
	Print("Duplicate receipt=%d:%d:%d: flags=0x%x, size=%d, limit=%d, total=%d.\n", pkt->pkt_xmit, pkt->pkt_bun, pkt->pkt_seq, pkt->pkt_flags, pkt->pkt_size, size, total);
#endif
}

static inline void
dbugReceiveAckDuplicate(
	struct packet *pkt,
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("Duplicate ACK=%d: expected=%d.\n", ack_pkt->pkt_xmit, pkt->pkt_xmit);
#endif
}

static inline void
dbugReceiveFinUnexpected(
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("Unexpected FIN: nacks=0x%x.\n", ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugReceiveCopy(
	int npacket, 
	int nbundle, 
	struct buffer *buff, 
	struct packet *pkt)
{
#ifdef DEBUGXXX
	Print("Copying DATA: index=%d, bundle=%d:%d, buff=0x%x, buff=0x%x\n", buff->index, nbundle, pkt->pkt_seq, buff->u, buff);
#endif
}

static inline void
dbugReceiveCheck(
	int count, 
	int window, 
	int retrans)
{
#ifdef DEBUG
	if (retrans)
		Print("Checking for retransmissions: count=%d, window=%d.\n", count, window);
#endif
}

static inline void
dbugResizeWindow(
	int count, 
	int window)
{
#ifdef DEBUGXXX
	Print("Resized window: count=%d, window=%d.\n", count, window);
#endif
}

static inline void
dbugPacketSize(
	int oddsize, 
	int count, 
	int last)
{
#ifdef DEBUGXXX
	Print("Packet oddsize: oddsize=%d, count=%d, last=%d.\n", oddsize, count, last); 
#endif
}

static inline void
dbugSendPrepRetrans(
	int window, 
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("Preparing to retransmit: old_win=%d, new_win=%d, nacks=0x%x.\n", window, ack_pkt->pkt_win, ack_pkt->pkt_nacks); 
#endif
}

static inline void
dbugReceivePrepRetrans(
	int window, 
	int new_window, 
	struct packet *ack_pkt)
{
#ifdef DEBUG
	Print("Preparing for retransmissions: old_win=%d, new_win=%d, nacks=0x%x.\n", window, new_window, ack_pkt->pkt_nacks);
#endif
}

static inline void
dbugUnreachable(
	char *hostname)
{
#ifdef DEBUG
	Print("Host unreachable.\n", hostname);
#endif
}
